/* header file for secsipid library */

#ifndef __SECSIPID_H__
#define __SECSIPID_H__

#include "libsecsipid.h"

/**
 * SECSIPID_VERSION = 0xAABBCC00
 * - corresponds to version AA.BB.CC
 * - last two hexdigits are left for patch extensions
 * - example: 0x1010000 corresponds to version 1.1.0
 */
#define SECSIPID_VERSION 0x1010000

#endif
